/** @typedef {-1 | 0 | 1} Ternary */
/**
 * @template T
 * @typedef {(a: T, b: T) => Ternary | NaN} Comparer
 */

/**
 * @template X, Y
 * @param {X | Y} a
 * @param {Y} b
 * @returns [X | undefined, Y | undefined]
 */
const optionalSecond = (a, b) => {
    let x = a
    let y = b

    if (!b) {
        x = undefined
        y = b
    }

    return [x, y]
}

/**
 * @template K, V
 */
export class OrderedMap extends Map {
    _sort

    /**
    * @overload
    * @param {Iterable.<[K, V]>} iterable
    * @param {Comparer<[K, V]>} comparer
    */
    /**
    * @overload
    * @param {Comparer<[K, V]>} comparer
    */
    /**
    * @param {Iterable.<[K, V]> | Comparer<[K, V]>} iterable
    * @param {Comparer<[K, V]>} [comparer]
    */
    constructor(iterable, comparer) {
        const [iter, sort] = optionalSecond(iterable, comparer)

        /** @ts-ignore */
        super(iter)
        this._sort = sort
    }

    /**
     * @param {Comparer<[K, V]>} comparer
     */
    sortWith(comparer) {
        this._sort = comparer
    }

    *[Symbol.iterator]() {
        /** @ts-ignore */
        yield* [...this.entries()].sort(this._sort)
    }
}

/**
 * @template T
 * @param {(value: T) => number} get
 * @returns {Comparer<T>}
 */
const prioritySort = get => (a, b) => get(a) - get(b)

/**
 * @template K, V
 */
export class PriorityMap extends OrderedMap {
    /**
     * @typedef {(value: [K, V]) => number} PriorityGetter
     */

    /** @type PriorityGetter */
    _getter

    /**
    * @overload
    * @param {Iterable.<[K, V]>} iterable
    * @param {PriorityGetter} getter
    */
    /**
    * @overload
    * @param {PriorityGetter} getter
    */
    /**
    * @param {Iterable.<[K, V]> | PriorityGetter} iterable
    * @param {PriorityGetter} [getter]
    */
    constructor(iterable, getter) {
        const [iter, get] = optionalSecond(iterable, getter)

        /** @ts-ignore */
        super(iter, prioritySort(get))

        /** @ts-ignore */
        this._getter = get
    }
}
